'use strict';

function base_canvas(canvas_id, opts) {

    var logo = {
        timestamp: 0,
        image: undefined,
    };


    function set_logo (img) {
        logo.image = img;
        logo.timestamp = Date.now();
    };

    function show_logo() {
        var img = new Image();
        img.onload = function () { set_logo(img); };
        img.src = '/img/anritsu-logo-dark.png';
    };

    show_logo();

    var min = Math.min;
    var max = Math.max;
    function round(x) {
        return 0.5 + Math.round(x);
    }

    function detect_change() {
        var initial = true;
        var key;
        return function did_change(val) {
            var this_key = JSON.stringify(val);
            if (initial || this_key !== key) {
                initial = false;
                key = this_key;
                return true;
            }
            else {
                return false;
            }
        };
    }

    opts = obj_extend(opts, {
        trace_bg_color: 'black',
        bg_color: '#231f20',
        bottom_font_color: '#fff',
        grid_color: '#aaa',
        trace_color: '#ff0',
        crosshair_color: '#289b7d',
        amplitude_units: 'dBm',
        font_size: 14,
        label_pad: 5,
        num_divs: 10,
        limit_color: 'rgb(0, 177, 235)',
        active_limit_color: 'rgb(98, 187, 70)',
    });

    var element= document.getElementById(canvas_id);

    var layers = {
        metadata: layer(),
        logo: layer(),
        model: layer(),
        name: layer(),
        unit_in_use_message:layer(),
        gps: layer(),
        time: layer(),
        messages: layer(),
    };

    function draw_metadata(ctx, metadata) {
        var placeholder = '--';
        ctx.fillStyle = opts.bottom_font_color;
        ctx.textBaseline = 'top';

        var x = opts.label_pad;
        ctx.font = styling(opts).bold_font();
        var em = ctx.measureText('M').width;

        var i = 0;
        var y = 0;
        var w = ctx.width - x;
        for (var k in metadata) {
            ctx.font = styling(opts).bold_font();
            ctx.textAlign = 'left';
            ctx.fillText(k, x, y, w);

            var vals = (metadata[k].constructor === Array) ?
                        metadata[k] : [metadata[k]];

            ctx.font = styling(opts).normal_font();
            for (var m = 0; m < vals.length; m++) {
                y += 1.5 * em;
                ctx.fillText(vals[m] || placeholder, x, y, w);
            }

            y += opts.label_pad + 2 * em;
        }
    }

    function draw_gps(ctx, text) {
        ctx.fillStyle = opts.bottom_font_color;
        ctx.font = styling(opts).normal_font();
        ctx.textAlign = 'right';
        ctx.textBaseline = 'baseline';
        var x = ctx.width;
        var y = ctx.height - opts.label_pad;
        text = text.replace(/&deg;/g, '\u00b0');
        ctx.fillText(text, x, y, ctx.width);
    }

    function draw_time(ctx, text) {
        ctx.fillStyle = opts.bottom_font_color;
        ctx.font = styling(opts).normal_font();
        ctx.textAlign = 'right';
        ctx.textBaseline = 'top';
        var x = ctx.width;
        var y = opts.label_pad;
        ctx.fillText(text, x, y, ctx.width);
    }

    function draw_messages(ctx, messages) {
        ctx.font = styling(opts).normal_font();
        ctx.textAlign = 'right';
        ctx.textBaseline = 'top';
        var x = ctx.width - opts.label_pad;

        for (var i = 0; i < messages.length; i++) {
            var label = messages[i].label;
            ctx.fillStyle = messages[i].style || opts.grid_color;
            var y = i * (1.5 * opts.font_size) + opts.label_pad;
            var w = ctx.width - 2 * opts.label_pad;
            styling(opts).outlined_text(ctx, label, x, y, w);
        }
    };

    var should_resize = detect_change();


    function resize_common_elements(display) {
        var pad = display.pad;
        var pad2 = display.pad2;
        var width = display.width;
        var height = display.height;

        {
            var x = 0;
            var y = 0;
            if (typeof logo.image !== 'undefined') {
                var nh = logo.image.naturalHeight;
                var nw = logo.image.naturalWidth;
                var h = Math.max(height / 8, 50);
                var w = h * nw / nh;
            }
            else {
                var w = 80;
                var h = 50;
            }
            if (display.top) {
                layers.logo.resize(x, y, w, h);
            }
            else {
                layers.logo.resize(x, y, 1, 1);
            }
        }

        {
            var ctx = layers.metadata.ctx;
            ctx.save();
            ctx.font = styling(opts).bold_font();
            var em = ctx.measureText('M').width;
            ctx.restore();

            var meta_width = 5 * em + pad2;
            var x = pad + meta_width / 4;
            var y = pad2 + layers.logo.rect().bottom();
            var w = meta_width + pad2 + x;
            var h = height - pad2;
            if (display.metadata) {
                layers.metadata.resize(x, y, w, h);
            }
            else {
                layers.metadata.resize(x, y, 1, 1);
            }
        }

        {
            var graphics_area = display.graphics_area();
            var x = layers.logo.rect().right();
            var y = layers.logo.rect().y;
            var w = (graphics_area.right - x) / 2;
            var h = layers.logo.rect().height;
            if (display.top) {
                layers.model.resize(x, y, w, h);
            }
            else {
                layers.model.resize(x, y, 0, 0);
            }
        }
        {
            var graphics_area = display.graphics_area();
            var x = layers.model.rect().x;
            var y = layers.model.rect().bottom()/4;
            var w = layers.model.rect().width;
            var h = layers.model.rect().height;
            layers.name.resize(x,y,w,h);
        }

        {
            var graphics_area = display.graphics_area();
            var x = layers.model.rect().x;
            var y = layers.model.rect().bottom()/2;
            var w = layers.model.rect().width;
            var h = layers.model.rect().height;
            layers.unit_in_use_message.resize(x,y,w,h);
        }

        {
            var graphics_area = display.graphics_area();
            var x = layers.model.rect().right();
            var y = layers.logo.rect().y;
            var w = graphics_area.right - x;
            var h = (layers.logo.rect().height / 2) || (opts.font_size * 1.2);
            layers.gps.resize(x, y, w, h);
        }

        {
            var x = layers.gps.rect().x;
            var y = layers.gps.rect().bottom();
            var w = layers.gps.rect().width;
            var h = (layers.logo.rect().height / 2) || (opts.font_size * 1.2);
            layers.time.resize(x, y, w, h);
        }

        {
            var x = display.graphics_area().x + pad;
            var y = display.graphics_area().y + pad;
            var w = display.graphics_area().w - pad2;
            var h = display.graphics_area().h - pad2
            layers.messages.resize(x, y, w, h);
        }
    }

    function draw_model_number(ctx, model) {
        var font_size = layers.logo.rect().height / 4;
        ctx.fillStyle = opts.bottom_font_color;
        ctx.font = font_size + 'px sans-serif';
        ctx.textAlign = 'left';
        ctx.textBaseline = 'middle';
        var x = 0;
        var y = ctx.height / 2;
        ctx.fillText(model, x, y, ctx.width);
    }

    function draw_unit_name(ctx,unit_name) {
        var font_size = layers.model.rect().height/8;
        ctx.fillStyle = opts.bottom_font_color;
        ctx.font = font_size + 'px sans-serif';
        ctx.textAlign = 'left';
        ctx.textBaseline = 'middle';
        var x = 0;
        var y = ctx.height/2;
        ctx.fillText(unit_name, x, y, ctx.width);
    }

    function draw_unit_in_use_message(ctx,message) {
        var font_size = layers.model.rect().height/8;
        ctx.fillStyle = "red";
        ctx.font = font_size + 'px sans-serif';
        ctx.textAlign = 'left';
        ctx.textBaseline = 'middle';
        var x = 0;
        var y = ctx.height/2;
        ctx.fillText(message, x, y, ctx.width);
    }

    function compose_common_elements(ctx){
        layers.metadata.compose_onto(ctx);
        layers.logo.compose_onto(ctx);
        layers.model.compose_onto(ctx);
        layers.name.compose_onto(ctx);
        layers.unit_in_use_message.compose_onto(ctx);
        layers.gps.compose_onto(ctx);
        layers.time.compose_onto(ctx);
        layers.messages.compose_onto(ctx);
    }


    function draw_common_elements(args) {

        if (args.font_size) {
            opts.font_size = parseInt(args.font_size);
        }

        if (element.height == 0 || element.width == 0) {
            return;
        }

        layers.metadata.draw_if_dirty(args.metadata, function (ctx) {
            draw_metadata(ctx, args.metadata);
        });

        layers.model.draw_if_dirty(args.model_number, function (ctx) {
            draw_model_number(ctx, args.model_number);
        });

        layers.name.draw_if_dirty(args.unit_name, function (ctx) {
            draw_unit_name(ctx, args.unit_name);
        });

        layers.unit_in_use_message.draw_if_dirty(args.unit_in_use_message, function (ctx) {
            draw_unit_in_use_message(ctx, args.unit_in_use_message);
        });

        layers.gps.draw_if_dirty(args.gps, function (ctx) {
            draw_gps(ctx, args.gps);
        });

        layers.time.draw_if_dirty(args.time, function (ctx) {
            draw_time(ctx, args.time);
        });

        if (typeof logo.image !== 'undefined') {
            layers.logo.draw_if_dirty(logo.timestamp, function (ctx) {
                var sx = 0;
                var sy = 0;
                var sWidth = logo.image.naturalWidth;
                var sHeight = logo.image.naturalHeight;
                var dx = 0;
                var dy = 0;
                var dWidth = ctx.width;
                var dHeight = ctx.height;
                // Check aspect ratio
                if (dWidth / dHeight - sWidth / sHeight > 0.01) {
                    console.warn('Logo has bad aspect ratio!');
                }
                ctx.drawImage(logo.image, sx, sy, sWidth, sHeight,
                                          dx, dy, dWidth, dHeight);
            });
        }

        layers.messages.draw_if_dirty(args.messages, function (ctx) {
            draw_messages(ctx, args.messages);
        });
    }

    var members = {
        layers: layers,
        element: element,
        _draw_other_elements: function(args){},
        _compose_other_elements: function(ctx){},
        _resize_other_elements: function(display){},
        draw: function(args){
            args = obj_extend(args, {
                points: [],
                messages: [],
                freq_labels: {
                    start: '',
                    stop: '',
                    center: '',
                    span: ''
                },
                metadata: {},
                model_number: '',
                unit_name:'',
                unit_in_use_message:'',
                gps: 'No GPS Fix',
                time: format_datetime(new Date()),
                display: { spectrum: true, spectrogram: true, },
            });

            var size_depends_on = [
                element.clientWidth,
                element.clientHeight,
                args.display,
                logo.timestamp,
            ];

            if (should_resize(size_depends_on)) {
                var display = args.display;
                display.pad = opts.label_pad;
                display.pad2 = 2 * display.pad;

                display.width = element.clientWidth;
                display.height = element.clientHeight;
                element.height = display.height;
                element.width = display.width;

                display.graphics_area = function(){
                    var x = layers.metadata.rect().right();
                    var y = layers.metadata.rect().y;
                    var w = display.width - x - display.pad - layers.metadata.rect().width / 2;
                    return {
                        x: x,
                        y: y,
                        w: w,
                        h: display.height - layers.logo.rect().height - 2*display.pad2,
                        right: x + w
                    };
                };

                resize_common_elements(display);
                members._resize_other_elements(display);
            }

            draw_common_elements(args);
            members._draw_other_elements(args);

            // compose the layers
            var ctx = element.getContext('2d');
            ctx.fillStyle = opts.bg_color;
            ctx.fillRect(0, 0, element.width, element.height);
            members._compose_other_elements(ctx);
            compose_common_elements(ctx);
        },
        set_cursor: function() {},
    };

    return members;
}
