// In Gecko browsers, drawing outlined text is relatively expensive. So,
// cache the layers. The grid and the top layer both contain outlined
// text and will not change frequently. The middle layer will change every
// frame.
function layer() {
    var last_key;
    var canvas = document.createElement('canvas');
    var ctx = canvas.getContext('2d');
    var x = 0, y = 0;
    var w = 0, h = 0;
    function disabled() { return (w < 1) || (h < 1); }
    return {
        draw: function (callback) {
            if (disabled()) return;
            ctx.save();
            ctx.clearRect(0, 0, w, h);
            callback(ctx);
            ctx.restore();
        },

        draw_if_dirty: function(value, callback) {
            if (disabled()) return;
            if (typeof value === 'function' &&
                typeof callback === 'undefined') {
                callback = value;
                value = '';
            }

            var key = JSON.stringify([value, canvas.width, canvas.height]);
            if (key !== last_key) {
                this.draw(callback);
            }
            last_key = key;
        },

        resize: function (new_x, new_y, new_w, new_h) {
            w = Math.round(new_w);
            h = Math.round(new_h);

            x = Math.round(new_x);
            y = Math.round(new_y);

            if (disabled()) return;

            canvas.width = w;
            canvas.height = h;
            ctx.width = w;
            ctx.height = h;

            ctx.clearRect(0, 0, w, h);
            last_key = undefined;
        },

        rect: function () {
            return {
                x: x,
                y: y,
                width: w,
                height: h,
                right: function () { return x + canvas.width; },
                bottom: function () { return y + canvas.height; },
            };
        },

        compose_onto: function(dest_ctx) {
            if (disabled()) return;
            dest_ctx.drawImage(canvas, x, y);
        },

        ctx: ctx,
        resizeLogic: function( stuff){},
    }
}