'use strict';

function canvas_spectrum(canvas_id, opts) {
    // HACK: really nasty copy paste
    opts = obj_extend(opts, {
        trace_bg_color: 'black',
        bg_color: '#231f20',
        bottom_font_color: '#fff',
        grid_color: '#aaa',
        trace_color: '#ff0',
        crosshair_color: '#289b7d',
        amplitude_units: 'dBm',
        font_size: 14,
        label_pad: 5,
        num_divs: 10,
        limit_color: 'rgb(0, 177, 235)',
        active_limit_color: 'rgb(98, 187, 70)',
    });

    var base = base_canvas(canvas_id, opts);

    // add new layers
    base.layers = obj_extend(base.layers,  {
        grid: layer(),
        trace: layer(),
        trace_frame: layer(),
        limits: layer(),
        markers: layer(),
        labels: layer(),
        cursor: layer(),
        spectrogram: layer(),
        waterfall_legend: layer(),
    });

    base.num_divs= function () {
        return opts.num_divs;
    };
    base.spectrogram_ctx =  function () {
        return layers.spectrogram.ctx;
    };

    base.mouse= function () {
        var width = base.layers.trace.rect().width;
        var height = base.layers.trace.rect().height;

        return {
            x: base.mouse_x / width,
            y: 1 - (base.mouse_y / height)
        };
    }


    var mouse_cursor = 'auto';
    var mouse_x = base.mouse_x;
    var mouse_y = base.mouse_y;
    var $canvas = $(document.getElementById(canvas_id));
    var crosshair_path = '/img/pencil.png';
    $.get(crosshair_path, function () { /* precache the image */});

    function update_mouse_cursor() {
        var tr = base.layers.trace.rect();
        if (base.mouse_x >= 0 && base.mouse_x < tr.width &&
            base.mouse_y >= 0 && base.mouse_y < tr.height) {
            $canvas.css('cursor', mouse_cursor);
        }
        else {
            $canvas.css('cursor', 'auto');
        }
    }
    $canvas.on('mousemove mouseup mousedown', function (event) {
        var offset_x = $(this).offset().left - $(window).scrollLeft();
        var offset_y = $(this).offset().top - $(window).scrollTop();

        offset_x += base.layers.trace.rect().x;
        offset_y += base.layers.trace.rect().y;

        base.mouse_x = Math.round(event.clientX - offset_x);
        base.mouse_y = Math.round(event.clientY - offset_y);

        update_mouse_cursor();
    });

    base.bounded_mouse= function () {
        if (!base.mouse_x || !base.mouse_y ||
            base.mouse_x < 0 || base.mouse_x > base.layers.trace.rect().width) {
            return null;
        }
        else {
            return base.mouse();
        }
    }


    $canvas.on('mouseup mousedown', update_mouse_cursor);

    function draw_cursor(ctx, ampl, freq, relative_x) {
        if (typeof relative_x !== 'undefined') {
            ctx.fillStyle = opts.crosshair_color;
            ctx.strokeStyle = opts.crosshair_color;
            var cursor_x_px = ctx.width * relative_x;
            ctx.beginPath();
            ctx.moveTo(cursor_x_px, 0);
            ctx.lineTo(cursor_x_px, ctx.height);
            ctx.stroke();
            ctx.globalAlpha = 1;

            ctx.font = styling(opts).normal_font();
            ctx.textBaseline = 'bottom';
            ctx.textAlign = 'right';

            var x = ctx.width - opts.label_pad;
            var y = ctx.height - opts.label_pad;

            styling(opts).outlined_text(ctx, freq, x, y);
            styling(opts).outlined_text(ctx, ampl, x,
                          y - 1.2 * opts.font_size - opts.label_pad);
        }
    }

    function draw_ylabels(ctx, ylabels) {
        ylabels = ylabels || ['???', '???', '???'];
        var line_height = 1.2 * opts.font_size + opts.label_pad;

        ctx.font = styling(opts).normal_font();
        ctx.fillStyle = opts.grid_color;

        var x = opts.label_pad;

        if (ylabels.length * line_height > ctx.height) {
            if (ylabels.length > 0) {
                ctx.textBaseline = 'top';
                var y = opts.label_pad;
                styling(opts).outlined_text(ctx, ylabels[0], x, y);
            }
            if (ylabels.length > 1) {
                ctx.textBaseline = 'bottom';
                var y = ctx.height - opts.label_pad;
                var txt = ylabels[ylabels.length - 1];
                styling(opts).outlined_text(ctx, txt, x, y);
            }
        }
        else {
            ctx.textBaseline = 'top';
            for (var i = 0; i < ylabels.length; i++) {
                var y = Math.round(i * ctx.height / ylabels.length);
                styling(opts).outlined_text(ctx, ylabels[i], x, y + opts.label_pad);
            }
        }
    }

    function draw_frame(ctx) {
        ctx.fillStyle = opts.grid_color;
        ctx.fillRect(0, 0, ctx.width, 1);
        ctx.fillRect(0, 0, 1, ctx.height);
        ctx.fillRect(0, ctx.height - 1, ctx.width, 1);
        ctx.fillRect(ctx.width - 1, 0, 1, ctx.height);
    }

    function draw_grid(ctx) {
        var xdivs = opts.num_divs;
        var ydivs = xdivs;
        var xdiv_width = ctx.width / xdivs;
        xdivs = 2 * Math.round(xdivs / 2);

        for (var i = 1; i < xdivs; i++) {
            var x = Math.floor(i * ctx.width / xdivs);

            ctx.fillStyle = opts.grid_color;
            ctx.fillRect(x, 0, 1, ctx.height);
        }

        for (var i = 1; i < ydivs; i++) {
            var y = Math.round(i * ctx.height / ydivs);

            ctx.fillStyle = opts.grid_color;
            ctx.fillRect(0, y, ctx.width, 1);
        }
    }

    function fit_labels(ctx, labels) {

        ctx.font = normal_font(1.5);
        var start_width = ctx.measureText(labels.start);
        var stop_width = ctx.measureText(labels.stop);

        ctx.font = normal_font();
        var middle_width = Math.max(
                ctx.measureText('Center ' + labels.center),
                ctx.measureText('Span ' + labels.span));

    }

    function draw_freq_labels(ctx, labels) {
        if (!ctx.width) return;
        ctx.fillStyle = opts.bottom_font_color;
        ctx.textBaseline = 'middle';

        var w = ctx.width / 3 - 2 * opts.label_pad;
        w = Math.max(1, w);

        /* center and span labels */ {
            ctx.textAlign = 'center';
            ctx.font = styling(opts).normal_font();
            var x = ctx.width / 2;

            {
                var y = ctx.height * 1 / 3;
                ctx.fillText('Center ' + labels.center, x, y, w);
            }

            {
                var y = ctx.height * 2 / 3;
                ctx.fillText('Span ' + labels.span, x, y, w);
            }
        }

        /* left and right labels */ {
            ctx.font = styling(opts).normal_font(1.5);
            var y = ctx.height / 2;

            {
                var x = 0;
                ctx.textAlign = 'left';
                ctx.fillText(labels.start, x, y, w);
            }

            {
                var x = ctx.width - opts.label_pad;
                ctx.textAlign = 'right';
                ctx.fillText(labels.stop, x, y, w);
            }
        }
    }

    function amplitude_to_pixels(amplitude, height) {
        var y = Math.max(0, Math.min(1, amplitude));
        // This is a bit of a fudge factor to draw a flat line at the
        // bottom of the trace
        return Math.max(1, Math.min(height - 2, Math.round(height * (1-y))));
    }

    function draw_markers(ctx, markers) {
        ctx.font = styling(opts).normal_font();
        var text_width = ctx.measureText('M').width;
        var text_height = text_width;

        (markers || []).map(function (marker) {
            var x = marker.freq * ctx.width;
            var y = amplitude_to_pixels(marker.amp, ctx.height);
            var text = marker.number.toString();

            // Set the size and position of the marker number box
            var box = {
                height: text_height + opts.label_pad,
                width: text_width + opts.label_pad,
                vert_offset: 7,
                horiz_offset: 5,
                x1: function() { return x + this.horiz_offset; },
                y1: function() { return y - this.vert_offset - this.height; },
            };

            // Draw the little circle on the trace
            ctx.strokeStyle = marker.style;
            ctx.lineWidth = 2;
            ctx.beginPath();
            ctx.arc(x, y, 2, 0, 2 * Math.PI);
            ctx.stroke();

            // Draw the background of the marker label
            ctx.fillStyle = marker.style;
            ctx.beginPath();
            ctx.fillRect(box.x1(), box.y1(), box.width, box.height);
            ctx.stroke();

            // Draw the text of the marker label
            ctx.fillStyle = opts.trace_bg_color;
            ctx.textAlign = 'center';
            ctx.textBaseline = 'middle';
            ctx.fillText(text, box.x1() + box.width/2,
                               box.y1() + box.height/2 + 1);
        });
    }

    function draw_trace(ctx, amplitudes, alphas, colors) {
        var width = ctx.width;
        var height = ctx.height;
        var round = Math.round;
        var len = amplitudes.length;

        var points = amplitudes.map(function (ampl, i) {
            // Snapping to half-integers gives the sharpest lines
            var ret = {
                x: 0.5 + round(width * i / (len-1)),
                y: 0.5 + round(amplitude_to_pixels(ampl, height)),
                alpha: alphas[i] || 1.0,
                color: colors[i] || opts.trace_color
            };
            ret.valid = !isNaN(ret.x) && !isNaN(ret.y);
            return ret;
        });

        ctx.lineWidth = 1;
        ctx.lineJoin = 'bevel';
        ctx.beginPath();

        for (var i = 0; i < len; i++) {
            var prev = points[i-1] || { valid: false };
            var point = points[i];

            if (point.color !== prev.color || point.alpha !== prev.alpha) {
                ctx.stroke();
                ctx.globalAlpha = point.alpha;
                ctx.strokeStyle = point.color;
                ctx.beginPath();

                // Prevent gaps when the color changes and the width is more
                // than the number of points.
                if (prev.valid) {
                    ctx.moveTo(prev.x, prev.y);
                }
            }

            if (point.valid) {
                if (prev.valid)
                    ctx.lineTo(point.x, point.y);
                else
                    ctx.moveTo(point.x, point.y);
            }
        };

        ctx.stroke();
    }

    function draw_limits(ctx, drawables) {
        var radius = 2;
        for (var i = 0; i < drawables.segments.length; i++) {
            if (i === drawables.active_index) {
                continue;
            }
            draw_one_limit(i, opts.limit_color);
        }

        radius = 4;
        // Make sure to draw the active one on top
        draw_one_limit(drawables.active_index, opts.active_limit_color);

        function draw_circle(x, y, radius) {
            ctx.beginPath();
            ctx.arc(x, y, radius, 0, 2 * Math.PI);
            ctx.stroke();
        }

        function draw_one_limit(i, color) {
            var segment = drawables.segments[i];
            if (typeof segment === 'undefined') return;
            var x1 = Math.round(ctx.width * segment.start);
            var x2 = Math.round(ctx.width * segment.stop);
            var y = Math.round(ctx.height * (1 - segment.amp));

            ctx.strokeStyle = color;
            ctx.lineWidth = 2;
            ctx.beginPath();
            ctx.moveTo(x1, y);
            ctx.lineTo(x2, y);
            ctx.stroke();

            draw_circle(x1, y, radius);
            draw_circle(x2, y, radius);
        }
    }

    base._draw_other_elements = function(args){

        args = obj_extend(args, {
            points: [],
            limit_lines: { active_index: -1, segments: [] },
            markers: [],
            ylabels: [],
            messages: [],
            cursor: { x: parseFloat(), freq: '', ampl: '' },
            freq_labels: {
                start: '',
                stop: '',
                center: '',
                span: ''
            },
            metadata: {},
            spectrogram: { draw: function (ctx) {} },
            model_number: '',
            unit_name:'',
            unit_in_use_message:'',
            gps: 'No GPS Fix',
            time: format_datetime(new Date()),
            display: { spectrum: true, spectrogram: true, },
        });

        base.layers.grid.draw_if_dirty(args.ylabels, function (ctx) {
            ctx.fillStyle = opts.trace_bg_color;
            ctx.fillRect(0, 0, base.element.width, base.element.height);
            draw_grid(ctx);
            draw_ylabels(ctx, args.ylabels);
        });

        base.layers.limits.draw_if_dirty(args.limit_lines, function (ctx) {
            draw_limits(ctx, args.limit_lines);
        });

        base.layers.trace_frame.draw_if_dirty(draw_frame);

        // Always assume the trace is dirty because it's probably
        // similarly expensive to stringify as to just draw it.
        base.layers.trace.draw(function (ctx) {
            draw_trace(ctx, args.points, args.alphas, args.colors);
        });

        base.layers.markers.draw_if_dirty(args.markers, function (ctx) {
            draw_markers(ctx, args.markers);
        });

        var cursor_mouse_x = base.mouse_x;
        if (args.cursor && typeof args.cursor.x !== 'undefined') {
            cursor_mouse_x = args.cursor.x;
        }

        base.layers.cursor.draw_if_dirty([cursor_mouse_x, args.cursor],
        function (ctx) {
            if (args.cursor)
                draw_cursor(ctx, args.cursor.ampl, args.cursor.freq,
                            cursor_mouse_x);
        });

        base.layers.labels.draw_if_dirty(args.freq_labels, function (ctx) {
            draw_freq_labels(ctx, args.freq_labels);
        });


        base.layers.spectrogram.draw(function (ctx) {
            ctx.fillStyle = opts.trace_bg_color;
            ctx.fillRect(0, 0, base.element.width, base.element.height);
            args.spectrogram.draw(ctx);
            draw_frame(ctx);
        });

        base.layers.waterfall_legend.draw_if_dirty(function (ctx) {
            args.spectrogram.draw_colormap(ctx);
        });
    };

    base._resize_other_elements = function(display){
        var pad = display.pad;
        var pad2 = display.pad2;
        var width = display.width;
        var height = display.height;

        // was too lazy to add the base prefix to everything
        var layers = base.layers

        if (display.spectrogram) {
            var spectrogram_height = Math.min(200,
                (height - layers.logo.rect().bottom()) * 0.25);
            if (height < 400) spectrogram_height = 0;
        }
        else {
            var spectrogram_height = 0;
        }

        var legend_width = Math.round(0.5 * Math.pow(opts.font_size, 1.5));
        if (!display.spectrogram) {
            legend_width = 0;
        }

        {
            var x = layers.metadata.rect().right() + legend_width;
            if (!display.metadata) {
                x = pad2;
            }
            var w = width - x - pad2 - layers.metadata.rect().width / 2;
            var h = Math.max(0, spectrogram_height - pad);
            var y = height - h - pad - layers.logo.rect().height / 2;

            layers.spectrogram.resize(x, y, w, h);
        }

        {
            var x = layers.spectrogram.rect().x;
            var w = layers.spectrogram.rect().width;
            var h = 2 * 1.5 * opts.font_size;
            var y = layers.spectrogram.rect().y - h - pad;
            layers.labels.resize(x, y, w, h);
        }

        {
            var x = layers.spectrogram.rect().x;
            var y = layers.metadata.rect().y;
            var w = layers.spectrogram.rect().width;
            var h = layers.labels.rect().y - pad - 1 - y;

            layers.grid.resize(x, y, w, h);
            layers.limits.resize(x, y, w, h);
            layers.markers.resize(x, y, w, h);
            layers.trace.resize(x, y, w, h);
            layers.trace_frame.resize(x-1, y-1, w+2, h+2);
            layers.cursor.resize(x, y, w, h);
        }

        {
            var x = layers.metadata.rect().right() + pad;
            var y = layers.trace.rect().y + 1;
            var w = layers.trace.rect().x - x - 1;
            var h = layers.trace.rect().height - 2;
            layers.waterfall_legend.resize(x, y, w, h);
        }
    }

    base._compose_other_elements = function(ctx){
        base.layers.grid.compose_onto(ctx);
        base.layers.trace.compose_onto(ctx);
        base.layers.trace_frame.compose_onto(ctx);
        base.layers.limits.compose_onto(ctx);
        base.layers.markers.compose_onto(ctx);
        base.layers.labels.compose_onto(ctx);
        base.layers.cursor.compose_onto(ctx);
        base.layers.spectrogram.compose_onto(ctx);
        base.layers.waterfall_legend.compose_onto(ctx);
    }

    base.set_cursor = function (x) {
        mouse_cursor = x || 'auto';
        if (mouse_cursor === 'crosshair') {
            mouse_cursor = 'url(' + crosshair_path + '), auto';
        }
        update_mouse_cursor();
    };
    return base;
}

