function poll_json_local_storage(key, default_val) {
    var save_callback;
    var cache = {};

    if (typeof default_val === 'undefined') {
        default_val = {};
    }

    if (typeof window.localStorage === 'object') {
        for (var k in window.localStorage) {
            if (window.localStorage.hasOwnProperty(k)) {
                cache[k] = localStorage.getItem(k);
            }
        }

        setInterval(function () {
            if (typeof save_callback === 'function') {
                var obj = JSON.stringify(save_callback());

                // Prevent contention between two tabs. This tab should not
                // save to localStorage unless it was changed *in this tab*.
                if (obj === cache[key]) return;
                if (obj === window.localStorage.getItem(key)) return;
                window.localStorage.setItem(key, obj);
            }
        }, 1000);

        return {
            onrestore: function (callback) {
                var obj = default_val;
                if (window.localStorage.hasOwnProperty(key)) {
                    try {
                        obj = JSON.parse(window.localStorage.getItem(key))
                    }
                    catch (e) {
                    }
                }

                callback(obj);
            },
            onsave: function (callback) {
                save_callback = callback;
            },
        };
    }
    else {
        return {
            onrestore: function (callback) {
                if (typeof callback === 'function') {
                    callback(default_val);
                }
            },
            onsave: function () {}
        };
    }
}


