'use strict';
$(function main() {
    var websocket_proto = ('https:' == document.location.protocol ? 'wss://' : 'ws://');
    var websocket_port = location.port || ('https:' == document.location.protocol ? '443' : '80');

    var lighttpd_mod_websocket_base_url = websocket_proto + location.hostname + ':' + websocket_port;

    var websocket_factories = [
        function websockify_websocket_factory() {
            var websockify_url = websocket_proto + location.hostname + ':4002';
            return new WebSocket(websockify_url, 'binary');
        },
        function ipv4_lighttpd_mod_websocket_websocket_factory() {
            return new WebSocket(lighttpd_mod_websocket_base_url + '/json.ws');
        },
        function ipv6_lighttpd_mod_websocket_websocket_factory() {
            return new WebSocket(lighttpd_mod_websocket_base_url + '/json6.ws');
        },
    ];

    function status(msg) {
        console.log(msg);
    }

    var client = json_tcp_client(websocket_factories, status);
    var scpi = ScpiClient(client);
    scpi.client = client;
    var cal_messages = CalMessages(scpi);
    var refresh_rate_message = trace_refresh_rate_message();

    get_ethernet_startup_settings(scpi,client);
    get_current_ethernet_startup_settings(scpi,client);
    get_dns_startup_settings(scpi,client);

    var trace_data = TraceData(client);
    var iq_data = [];
    var iq_metadata = {};
    function set_iq_data(data){
        iq_data = data.map(function(sample){return [sample.i, sample.q];});
        if (data && data.length > 0) {
            iq_metadata = iq_metadata_snapshot(scpi);
        }
        else {
            iq_metadata = {};
        }
    }

    var spectrogram = Spectrogram.create(400);

    var spectrum = canvas_spectrum('spectrum');

    var limits = Limits(spectrum, trace_data, scpi);
    var markers = Markers(spectrum, trace_data);
    var trace_status = TraceStatus(trace_data);
    var save_on_event = SaveOnEvent(scpi);

    setup_y_scale_commands(scpi, trace_data, spectrum);
    setup_frequency_commands(scpi, trace_data);
    setup_overpower(scpi);
    init_get_metadata(trace_data);
    bind_setting_change_handlers(scpi);
    setup_scpi_console(client);
    setup_overheating_msg(client);
    setup_disconnect_msg(client);
    setup_amplitude_span(trace_data, spectrum);
    setup_reboot(scpi);
    setup_option_upgrade(client);
    setup_firmware_update(client, scpi);
    setup_dhcp(scpi);
    setup_ethernet(scpi);
    setup_dns(scpi);
    setup_show_firmware_info(scpi);
    setup_unit_name_info(scpi);
    setup_show_nfsmount_info(scpi)
    setup_bandwidth_ratio(scpi);
    setup_app_version(client);
    setup_storage_info(client);
    setup_report_bug(scpi);
    setup_debug_input(client, scpi);
    setup_scpi_script(scpi);
    var tabs = Tabs(scpi);
    setup_gps_status(client);
    setup_average_count(scpi);
    setup_obw_method(scpi);
    setup_rl_buttons(scpi);
    setup_display_filters();
    setup_span_up_down(scpi);
    setup_limit_checking(scpi);
    setup_iq_capture_result(client, set_iq_data);
    setup_file(client, scpi);
    setup_trace_refresh_rate(trace_data);

    var limit_alarm = LimitAlarm(client);

    var cursor = Cursor();

    function animate_spectrum(){
        var normalized = trace_data.normalized_trace_data();

        var alphas = trace_data.stale_mask().map(function (stale) {
            return stale ? 0.3 : 1;
        });

        var spectrum_mouse = spectrum.bounded_mouse();
        var frequency = trace_data.frequency;

        var idn = instrument_idn($("#firmware-info").text());
        var model_number = idn.model || '';

        var unit_name = $("#unit-name").text();
        var is_checked = $("input[name=unit-in-use-message]").prop('checked');

        var unit_in_use_message = is_checked ? "RSM IN USE" : "";

        var messages = trace_status.get_messages(trace_data).concat(
                      cal_messages.get()).concat(refresh_rate_message.get())

        if (limit_alarm.is_alarming()) {
            messages.push({
                label: 'Limit Failure',
                style: 'cyan',
            });
        }

        spectrum.draw({
            points: normalized,
            markers: markers.get_drawables(),
            limit_lines: limits.get_drawables(),
            alphas: alphas,
            colors: trace_status.get_colors(),
            ylabels: trace_data.get_ref_labels(10),
            cursor: cursor.get(spectrum_mouse, trace_data),
            freq_labels: {
                start: format_freq(frequency.start),
                stop: format_freq(frequency.stop),
                center: format_freq(frequency.center),
                span: format_freq(frequency.span)
            },
            messages: messages,
            font_size: font_size(),
            metadata: get_spectrum_metadata(scpi),
            spectrogram: spectrogram,
            model_number: model_number,
            unit_name:unit_name,
            unit_in_use_message:unit_in_use_message,
            gps: global_gps_description,
            display: {
                spectrogram: spectrogram_visible(),
                top: true,
                metadata: true
            },
        });
    };

    function animate_iq(){
        var idn = instrument_idn($("#firmware-info").text());
        var model_number = idn.model || '';

        spectrum.draw({
            points: iq_data,
            font_size: font_size(),
            metadata: iq_metadata,
            model_number: model_number,
            gps: global_gps_description,
        });
    };

    spectrum.animate = animate_spectrum;

    function display_spectrum(){
        spectrum = canvas_spectrum('spectrum');
        spectrum.animate = animate_spectrum;
        trace_data.unpause();
    };

    function display_iq(){
        trace_data.pause();
        spectrum = iq_canvas('spectrum');
        spectrum.animate = animate_iq;
    };

    tabs.onfocus('iq-capture', display_iq);
    tabs.onblur('iq-capture', display_spectrum);

    tabs.onfocus('limits', limits.focus);
    tabs.onblur('limits', limits.blur);

    tabs.onfocus('markers', markers.focus);
    tabs.onblur('markers', markers.blur);

    tabs.onfocus('markers', cursor.focus);
    tabs.onblur('markers', cursor.blur);

    var spectrum_crosshair_selectors = [
        '#marker-table .marker-freq',
        '.limit-table .limit-amp',
        '.limit-table .limit-freq'].join(', ');

    $('body').on('focus', spectrum_crosshair_selectors, function () {
        $('#spectrum').addClass('selecting');
    });

    $('body').on('blur', spectrum_crosshair_selectors, function () {
        $('#spectrum').removeClass('selecting');
    });


    (function () {
        window.requestAnimationFrame(function continue_animation(){
            spectrum.animate();
            window.requestAnimationFrame(continue_animation);
        } );
    })();

    (function () {
        var prev_id;
        trace_data.on_data(function () {
            var id = trace_data.id();
            if (id !== prev_id) {
                prev_id = id;
                spectrogram.shift();
            }
            spectrogram.new_trace_data(trace_data.normalized_trace_data())
        });
    })();

    scpi.on('init:cont', function (val) {
        var disabled = val == 1;
        $('[data-scpi="init:imm"]').attr('disabled', disabled);
    });
});
