function Markers(spectrum, trace_data) {
    var table = marker_table('marker-table', 8, trace_data);
    var dragging_row;

    function mouse_freq() {
        var mouse = spectrum.bounded_mouse();
        if (mouse) {
            var x = clamp(mouse.x, 0, 1);
            return trace_data.frequency.from_relative(x);
        }
    }

    var downTimer;

    function mousedown() {
        clearTimeout(downTimer);
        downTimer = setTimeout(function() {
            // Don't create a new marker if one is already active
            if (!dragging_row) {
                dragging_row = table.add_new_marker(mouse_freq());
            }
        }, 2000);

        dragging_row = table.maybe_place_marker(mouse_freq());
    }

    function mouseup() {
        // TODO should this clear the document.activeElement if it was
        // previously in the table row, or does "the user" expect it to
        // remain? Does the user want to click and drag multiple times on the
        // same one? or instead, click and drag multiple times to place
        // multiple markers?

        // The current behavior is that after placing a new marker, the next
        // click-drag will affect that new one. After that, the row is not
        // selected. If this behavior is changed, please update this comment
        // to reflect that fact.
        dragging_row = undefined;
        clearTimeout(downTimer);
    }

    function mousemove() {
        if (dragging_row) {
            dragging_row.set({
                freq: mouse_freq(),
                enabled: true
            });
        }
    }

    var handlers = {
        mousedown: mousedown,
        mouseup: mouseup,
        mousemove: mousemove,
    }

    function enable(e) {
        dragging_row = e ? dragging_row : undefined;

        for (var key in handlers) {
            $('#spectrum')[e ? 'on' : 'off'](key, handlers[key]);
        }
    }

    function to_relative(marker) {
        var active_num = dragging_row && dragging_row.get().number;
        var is_active = active_num === marker.number;

        return {
            freq: trace_data.frequency.to_relative(marker.freq),
            amp: trace_data.normalize(marker.amp),
            number: marker.number,
            style: is_active?'#D14B2A':'#7f7',
        };
    }

    function drawable(markers) {
        return markers.filter(function (marker) {
            return marker.enabled;
        }).map(to_relative);
    }

    return {
        get_drawables: function () {
            return drawable(table.get_markers());
        },

        table: table,

        focus: function () {
            enable(true);
        },
        blur: function () {
            enable(false);
        },
    };
}
