'use strict'

var snap = (function () {
    var threshold = 0.01;

    function snap(val, direction) {
        var sign = Math.sign(val);
        val *= sign;

        var order = Math.floor(Math.log10(val));
        var x = val / Math.pow(10, order);

        var choices = [0.5, 1, 2, 5, 10, 20];

        var choice;
        if (direction > 0) {
            for (choice = 0; choice < choices.length; choice++) {
                if (choices[choice] > x + threshold) {
                    break;
                }
            }
        }
        else {
            for (choice = choices.length - 1; choice >= 0; choice--) {
                if (choices[choice] < x - threshold) {
                    break;
                }
            }
        }

        return sign * choices[choice] * Math.pow(10, order);
    }

    (function snap_selftest() {
        function test(direction, items) {
            for (var i = 0; i < items.length; i++) {
                var actual = snap(items[i][0], direction);
                var expected = items[i][1];

                if (actual !== expected) {
                    console.error('Snap self test failed: snapping ' +
                        items[i][0] + (direction > 0 ? ' up' : ' down') +
                        ' is ' + actual + ' not ' + expected);
                }
            }
        }

        test(1, [
            [1, 2],
            [1.5, 2],
            [2, 5],
            [2.5, 5],
            [7, 10],
            [0.9, 1],
            [10, 20],
            [10.01, 20],
            [0.99, 1],
            [9.99, 20]
        ]);

        test(-1, [
            [1, 0.5],
            [1.5, 1],
            [2, 1],
            [2.5, 2],
            [7, 5],
            [0.9, 0.5],
            [10, 5],
            [10.01, 5]
        ]);
    })();

    return {
        up: function (x) { return snap(x, 1); },
        down: function (x) { return snap(x, -1); },
    };
})();
