'use strict';

function limit_table(spectrum, trace_data) {
    var $limit_table = $('#limit-table');
    var $tbody = $('#limit-table tbody');
    var $enabled = $('#limit-enabled');

    var $limit_mask = $('#limit-mask');
    var $apply_mask = $('#limit-apply-mask');
    var $max_hold = $('#limit-set-max-hold');
    var $min_hold = $('#limit-set-min-hold');
    var $mask_offset = $('#limit-mask-offset');
    var $save_on_event = $('#save-on-event');

    var $relative_freq = $('#limit-relative-freq');
    var $relative_amp = $('#limit-relative-amp');
    var $copy_button = $('#limit-copy-row');
    var $add_button = $('#limit-add-row');
    var $del_button = $('#limit-del-row');
    var $draw_button = $('#limit-draw-segment');
    var $split_button = $('#limit-split-segment');
    var $spectrum = $(spectrum.element);

    var $edit_start = $('#limit-edit-start-freq');
    var $edit_stop = $('#limit-edit-stop-freq');
    var $edit_amp = $('#limit-edit-amp');

    var $edit_limit = $('#limit-edit-start-freq','#limit-edit-stop-freq','#limit-edit-amp', '#limit-table tbody');
    $edit_limit.on('change', function() { $limit_table.trigger('change'); });

    function $rows() {
        return $tbody.find('.limit-table-row');
    }
    function is_relative_freq() {
        return $relative_freq.is(':checked');
    }
    function is_relative_amp() {
        return $relative_amp.is(':checked');
    }

    var model = {
        enabled: true,
        frequencyRelative: false,
        amplitudeRelative: false,
        segments: [],
    };

    function create_mask(isLower) {

        model.segments.forEach(function(segment, index, segments){
            if (segment.isLower === isLower) {
                segments.splice(index, 1);
            }
        });

        var start_frequency = trace_data.frequency.start;
        var stop_frequency = trace_data.frequency.stop;
        var segment_length_Hz = (stop_frequency - start_frequency)/40;

        var trace_value_pairs = trace_data.pairs();
        var num_of_trace_points = trace_value_pairs.length;
        var trace_points_in_segment = num_of_trace_points/40;

        var trace_point = 0;
        for (var current_frequency = start_frequency;
             current_frequency < stop_frequency;
             current_frequency = current_frequency + segment_length_Hz) {

            var segment_amplitudes = trace_value_pairs.slice(trace_point, trace_point + trace_points_in_segment).map(function (obj) {
                return obj.amp;
            });

            var peak_amplitude = isLower ? Math.min.apply(Math, segment_amplitudes) : Math.max.apply(Math, segment_amplitudes);
            trace_point += trace_points_in_segment;

            append_segment();
            active_segment_value({
                lower: isLower,
                start: current_frequency,
                stop: Math.min(current_frequency + segment_length_Hz, stop_frequency),
                amp: peak_amplitude + (isLower ? -parseInt($mask_offset.val()) : parseInt($mask_offset.val()))
            });
        }
    }

    $limit_mask.on('change', function() {
        var limitMask = $(this).is(':checked');
        if  (!limitMask) {

            //Clear the limits table.
            model.segments = [];
        }
        else
        {
            //Clear the limits table so a mask can be created.
            model.segments = [];

            create_mask(false);
            create_mask(true);
        }
        update_dom();
    });

    $save_on_event.on('change', function() {
        var $this = $(this);
        var isLimitLineSet = model.segments.length > 0;
        var isLimitCheckingEnabled = $enabled.prop('checked');

        if (!isLimitCheckingEnabled) {
            $this.prop('checked', false);
            var dialog = modal_overlay('#save-on-event-error-dialog');
            dialog.show();
        }
        else if (!isLimitLineSet) {
            $this.prop('checked', false);
            console.log('Limit line is not set.');
        }
    });

    var active_segment_index = -1;
    function append_segment() {
        model.segments.push({
            lower: false,
            start: parseFloat(),
            stop: parseFloat(),
            amp: parseFloat(),
        });
        active_segment_index = model.segments.length - 1;
        return active_segment_index;
    }

    function active_segment_value(val) {
        if (active_segment_index > -1) {
            return segment_value(active_segment_index, val);
        }
        else {
            return {};
        }
    }

    function segment_value(index, val) {
        if (index < 0) {
            throw 'Negative segment index';
        }

        var segment = model.segments[index];
        if (typeof segment === 'undefined') {
            throw 'no segment at index ' + index;
        }

        function maybe_set(obj, key, value) {
            if (typeof value !== 'undefined') {
                if (isNaN(+value)) {
                    throw 'Bad value';
                }
                obj[key] = value;
            }
        }

        if (typeof val !== 'undefined') {
            maybe_set(segment, 'lower', val.lower);
            maybe_set(segment, 'start', val.start);
            maybe_set(segment, 'stop', val.stop);
            maybe_set(segment, 'amp', val.amp);
        }

        return {
            lower: segment.lower,
            start: segment.start,
            stop: segment.stop,
            amp: segment.amp,
        };
    }

    call(function init_inputs() {
        function aux($el, inp_type, key, parse, format) {
            $el.setting_input('init', inp_type);
            $el.on('commit', function (_, raw_val) {
                var val = parse(raw_val);
                if (active_segment_index > -1) {
                    var arg = {};
                    arg[key] = val;
                    active_segment_value(arg);
                    update_dom();
                }
            });
        }

        aux($edit_start, 'frequency', 'start', parse_freq, format_freq);
        aux($edit_stop, 'frequency', 'stop', parse_freq, format_freq);
        aux($edit_amp, 'number', 'amp', parse_amp, format_amplitude);
    });

    var row_prototype_html = call(function init_limit_table() {
        var $prototype = $tbody.find('.limit-table-prototype:first');

        if ($prototype) {
            $tbody.find('.limit-table-prototype').remove();

            // This gets the outer HTML of the prototype since html()
            // returns the inner HTML only
            $prototype.removeClass('limit-table-prototype');
            $prototype.find().prop('disabled', false);
            return $('<div>').append($prototype.clone()).html();
        }
        else {
            console.error('No prototype-html for limit_table');
        }
    });

    function call(f) { return f(); }

    function mouse_coords_to_physical(x, y) {
        var center = is_relative_freq() ? trace_data.frequency.center : 0;
        var ref_level = is_relative_amp() ? trace_data.ref_level() : 0;

        return {
            freq: trace_data.frequency.from_relative(x) - center,
            amp: trace_data.denormalize(y) - ref_level,
        };
    }

    var set_drawing_mode = call(function () {
        var dragging = false;

        function update_position() {
            var mouse = spectrum.bounded_mouse();
            if (dragging && mouse) {
                var phys = mouse_coords_to_physical(mouse.x, mouse.y);
                active_segment_value({
                    stop: phys.freq,
                    amp: phys.amp,
                });
            }
        }

        var handlers = {
            mousedown: function () {
                var mouse = spectrum.bounded_mouse();
                if (mouse) {
                    append_segment();
                    dragging = true;
                    var phys = mouse_coords_to_physical(mouse.x, mouse.y);
                    active_segment_value({
                        start: phys.freq,
                        stop: phys.freq,
                        amp: phys.amp,
                    });
                }
                else {
                    enable(false);
                }
            },
            mousemove: function () {
                update_position();
            },
            mouseup: function () {
                update_position();
                enable(false);
                set_active_row(model.segments.length - 1);
                set_dragging_mode(true);
            },
        };

        function enable(enabled) {
            if (enabled) {
                spectrum.set_cursor('crosshair');
            }
            else {
                dragging = false;
                spectrum.set_cursor();
                update_dom();
            }

            var $spectrum = $(spectrum.element);
            for (var key in handlers) {
                $spectrum[enabled ? 'on' : 'off'](key, handlers[key]);
            }
        }

        return enable;
    });

    $add_button.on('click', function () {

        var span = trace_data.frequency.span;
        var center = is_relative_freq() ? 0 : trace_data.frequency.center;

        var start_frequency;
        var stop_frequency;
        var amp;
        var value;
        var units;

        // if the input fields are empty, use the current center and span setting
        //else use the input fields for start and stop frequency
        if($edit_start.val())
        {
             value = $edit_start.val();
             units = $edit_start.next().find('option:selected').text();
             start_frequency = parse_freq(value + ' ' + units);
        }
        else
        {
            start_frequency = center - span / 4;
        }

        if($edit_stop.val())
        {
            value = $edit_stop.val();
            units = $edit_stop.next().find('option:selected').text();
            stop_frequency = parse_freq(value + ' ' + units);
        }
        else
        {
            stop_frequency =  center + span / 4;
        }

        if($edit_amp.val())
        {
            amp = parse_amp($edit_amp.val());
        }
        else
        {
            amp = trace_data.denormalize(0.5);
        }

        append_segment();
        active_segment_value({
            lower: false,
            start: start_frequency,
            stop: stop_frequency,
            amp: amp
        });

        update_dom();
        set_active_row(model.segments.length - 1);

        set_drawing_mode(false);
        set_dragging_mode(true);
    });

    function copy_active_row() {
        var prev_index = active_segment_index;
        var prev_value = active_segment_value();

        if (prev_index > -1 && prev_value) {
            append_segment();
            active_segment_value(prev_value);

            var new_segment = model.segments.splice(active_segment_index, 1)[0];
            model.segments.splice(prev_index, 0, new_segment);

            set_drawing_mode(false);
            set_dragging_mode(true);
            update_dom();
            set_active_row(prev_index + 1);
        }
    }

    $enabled.on('change', function () {
        model.enabled = $enabled.is(':checked');
    });

    $copy_button.on('click', function () {

        copy_active_row();
    });

    $draw_button.on('click', function () {

        active_segment_index = -1;
        set_dragging_mode(false);
        set_drawing_mode(true);
    });

    $(document.body).on('mousedown', function () {
        if (!spectrum.bounded_mouse()) {
            set_drawing_mode(false);
            set_dragging_mode(true);
        }
    });

    $del_button.on('click', function () {

        if (active_segment_index > -1) {
            model.segments.splice(active_segment_index, 1);
            set_active_row(Math.min(
                model.segments.length - 1, active_segment_index));
            update_dom();
        }
    });

    $relative_freq.on('change', function () {
        var freq_offset;

        if ($(this).is(':checked')) {
            // from absolute to relative
            freq_offset = -trace_data.frequency.center;
        }
        else {
            freq_offset = trace_data.frequency.center;
        }

        function adjust(el) {
            var $el = $(el);
            $el.val(format_freq(parse_freq($el.val()) + freq_offset));
        }

        model.segments = model.segments.map(function (segment) {
            segment.start += freq_offset;
            segment.stop += freq_offset;
            return segment;
        });

        model.frequencyRelative = $relative_freq.is(':checked');
        update_dom();
    });

    $relative_amp.on('change', function () {
        var amp_offset;

        if ($(this).is(':checked')) {
            // from absolute to relative
            amp_offset = -trace_data.ref_level();
        }
        else {
            amp_offset = trace_data.ref_level();
        }

        model.segments = model.segments.map(function (segment) {
            segment.amp += amp_offset;
            return segment;
        });

        model.amplitudeRelative = $relative_amp.is(':checked');
        update_dom();
    });

    $split_button.on('click', function () {

        if (active_segment_index > -1) {
            var prev_value = active_segment_value();
            var prev_index = active_segment_index;

            var mid = (prev_value.start + prev_value.stop) / 2;

            copy_active_row();
            segment_value(prev_index, { stop: mid });
            active_segment_value({ start: mid });
            update_dom();
        }
    });

    function deserialize(obj) {
        /* Requires obj to be an object of the form:
         *  {
         *      enabled: true,
         *      frequencyRelative: true,
         *      amplitudeRelative: true,
         *      segments: [
         *          {
         *              frequency: { start: 123e4, stop 456e7 },
         *              amplitude: { unit: 'dBm', value: -100 },
         *          },
         *          ...
         *      ]
         *  }
         */
        if (typeof obj.enabled !== 'boolean' ||
            typeof obj.frequencyRelative !== 'boolean' ||
            typeof obj.amplitudeRelative !== 'boolean' ||
            !Array.isArray(obj.segments))
        {
            console.log(obj);
            throw 'Invalid limit object'
        }

        for (var i = 0; i < obj.segments.length; i++) {
            var seg = obj.segments[i];
            var freq = seg.frequency || {};
            var amp = seg.amplitude || {};
            if (typeof freq.start !== 'number' ||
                isNaN(freq.start) ||
                typeof freq.stop  !== 'number' ||
                isNaN(freq.stop) ||
                isNaN(amp.value) ||
                typeof amp.value  !== 'number' ||
                amp.unit.toUpperCase() !== trace_data.units().toUpperCase())
            {
                console.log(obj);
                throw 'Invalid limit object';
            }
        }

        $relative_freq.prop('checked', obj.frequencyRelative);
        $relative_amp.prop('checked', obj.amplitudeRelative);

        model = {
            enabled: obj.enabled,
            frequencyRelative: obj.frequencyRelative,
            amplitudeRelative: obj.amplitudeRelative,
            segments: obj.segments.map(function (segment) {
                return {
                    lower: segment.lower,
                    start: segment.frequency.start,
                    stop: segment.frequency.stop,
                    amp: segment.amplitude.value,
                };
            }),
        };

        update_dom();
    }

    function update_dom() {

      $relative_freq.prop('checked', model.frequencyRelative);
      $relative_amp.prop('checked', model.amplitudeRelative);

      if (model.segments.length === 0) {
          if ($rows().length != 0) {
                    $rows().remove();
                    $tbody.change();
            }
            return;
        }

        var missing_rows = model.segments.length - $rows().length;

        while (missing_rows > 0) {
            $tbody.append($(row_prototype_html));
            missing_rows--;
        }

        while (missing_rows < 0) {
            $tbody.find('.limit-table-row:last').remove();
            $tbody.change();
            missing_rows++;
        }

        var rows = $rows();

        for (var i = 0; i < model.segments.length; i++) {
            var seg = model.segments[i];
            var $row = $(rows[i]);
            limit_index($row, i);

            var $start = $row.find('.limit-start-freq').first();
            var $stop = $row.find('.limit-stop-freq').first();
            var $amp = $row.find('.limit-amp').first();
            var $bound = $row.find('.limit-bound').first();

            $bound.prop('checked', seg.lower);
	             var formatted_start = format_freq(seg.start);
            var formatted_stop = format_freq(seg.stop);
            var formatted_amplitude = format_amplitude(seg.amp, trace_data.units());

	          if (formatted_start != $start.val()) $start.val(formatted_start).change();
            if (formatted_stop != $stop.val()) $stop.val(formatted_stop).change();
            if (formatted_amplitude != $amp.val()) $amp.val(formatted_amplitude).change();
        }

        if (active_segment_index > -1) {
            var seg = active_segment_value();
            $edit_start.setting_input('value', seg.start);
            $edit_stop.setting_input('value', seg.stop);
            $edit_amp.setting_input('value', seg.amp);
        }
    }

    function limit_index(el, arg) {
        var $row = $(el).closest('.limit-table-row');
        // jQuery.data counts arguments...
        if (typeof arg === 'undefined')
            return $row.data('limit-index');
        else
            return $row.data('limit-index', arg);
    }

    $tbody.on('click', '.limit-table-row input', function () {
        set_active_row(limit_index(this));
        set_drawing_mode(false);
        set_dragging_mode(true);
        if ($(this).hasClass('limit-start-freq')) {
            $edit_start.focus().select();
        }
        else if ($(this).hasClass('limit-stop-freq')) {
            $edit_stop.focus().select();
        }
        else if ($(this).hasClass('limit-amp')) {
            $edit_amp.focus().select();
        }
        else if ($(this).hasClass('limit-bound')) {
            var $active_bound = get_active_row().find('.limit-bound').first();
            var new_value = !$active_bound.prop('checked');
            var segment = model.segments[active_segment_index];

            $active_bound.prop('checked', new_value);
            segment['lower'] = new_value;

        }
        else {
            console.error('What is this');
            console.log($(this));
        }
    });

    function get_active_row() {
        return get_row(active_segment_index);
    }

    function get_row(index) {
        var rs = $rows();
        for (var i = 0; i < rs.length; i++) {
            if (limit_index(rs[i]) === index) {
                return $(rs[i]);
            }
        }
        return $();
    }

    function set_active_row(index) {
        var $edit_elems = [$edit_start, $edit_stop, $edit_amp];
        $rows().removeClass('active-row');

        active_segment_index = index;
        var active_row = get_row(index);

        if (active_row.length > 0) {
            active_row.addClass('active-row');
            var vals = active_segment_value();

            $edit_start.setting_input('value', vals.start);
            $edit_stop.setting_input('value', vals.stop);
            $edit_amp.setting_input('value', vals.amp);
        }
        else {
            $edit_elems.map(function ($el) {
                $el.setting_input('value', '');
            });
            spectrum.set_cursor();
        }
    }

    var set_dragging_mode = call(function () {
        var dragging = false;
        var drag_x = parseFloat();
        var drag_y = parseFloat();
        var drag_from;

        function mousedown() {
            var m = spectrum.mouse();

            if (m.x >= 0 && m.x <= 1 &&
                m.y >= 0 && m.y <= 1 &&
                active_segment_index > -1
            ) {
                dragging = true;
                drag_x = m.x;
                drag_y = m.y;
                drag_from = active_segment_value();
            }
        }

        function mouseup() {
            mousemove();
            dragging = false;
            update_dom();
        }

        function mousemove() {
            if (dragging) {
                var m = spectrum.mouse();

                var dfreq = trace_data.frequency.from_relative(m.x) -
                            trace_data.frequency.from_relative(drag_x);
                var damp = trace_data.denormalize(m.y) -
                           trace_data.denormalize(drag_y);

                active_segment_value({
                    start: drag_from.start + dfreq,
                    stop: drag_from.stop + dfreq,
                    amp: drag_from.amp + damp
                });
            }
        }

        var handlers = {
            mousedown: mousedown,
            mousemove: mousemove,
            mouseup: mouseup,
        };

        var was_enabled = false;
        return function (enabled) {
            if (enabled) {
                spectrum.set_cursor('move');
            }
            else {
                spectrum.set_cursor();
            }

            if (was_enabled != enabled) {
                was_enabled = enabled;
                for (var key in handlers) {
                    $spectrum[enabled ? 'on' : 'off'](key, handlers[key]);
                }
            }
        }
    });

    var methods = {
        get: function () {
            return {
                amplitudeRelative: model.amplitudeRelative,
                frequencyRelative: model.frequencyRelative,
                enabled: model.enabled,
                segments: model.segments.map(function (segment) {
                    return {
                        lower: segment.lower,
                        frequency: {
                            start: segment.start,
                            stop: segment.stop,
                        },
                        amplitude: {
                            value: segment.amp,
                            unit: trace_data.units(),
                        },
                    };
                }),
            };
        },
        set: function (obj) {
            deserialize(obj);
            set_active_row();
        },
        initialize: function(obj) {
            deserialize(obj);
            set_active_row();
            $limit_table.trigger('initialized');
        },
        active_index: function () {
            return active_segment_index;
        },
        blur: function () {
            set_dragging_mode(false);
            set_drawing_mode(false);
            set_active_row();
        },
        focus: function () {
            set_dragging_mode(true);
            set_drawing_mode(false);
        },
    };

    update_dom();
    return methods;
}
