'use strict'

function marker_table(table_id, num_rows, trace_data) {
    var $table = $('#' + table_id);
    var rows = [];

    function call(f) { return f(); }

    function swap(obj, a, b) {
        var tmp = obj[a];
        obj[a] = obj[b];
        obj[b] = tmp;
    }

    function table_row(elem, number) {
        var $row = $(elem);
        var $amp = $row.find('.marker-amp');

        function clear() {
            if ($amp.text() !== '') {
                $amp.text('');
            }
        }

        function bool_input(selector) {
            var $input = $row.find(selector);

            var value = parseFloat();
            var text = '';

            $input.on('change', function () {
                value = $input.is(':checked');
                text = String(value);

                if (!value) {
                    clear();
                }
            });

            return {
                $: $input,
                text: function () { return text; },
                value: function () { return value; },
                set: function (new_val) {
                    $input.prop('checked', new_val);
                    $input.trigger('change');
                },
            }
        }

        function numeric_input(selector, parse) {
            var $input = $row.find(selector);
            var value = parseFloat();
            var text = '';

            $input.on('keyup change', function () {
                var new_val = parse($input.val());
                if (!isNaN(new_val)) {
                    text = $input.val();
                    value = new_val;
                    inputs.enabled.set(true);
                }
            });

            $input.on('blur', function () {
                if (isNaN(value)) {
                    text = '';
                    $input.val('');
                }
                else {
                    methods.set(value);
                }
            });

            var methods = {
                $: $input,
                text: function () { return text; },
                value: function () { return value; },
                set: function (new_val) {
                    $input.val(format_freq(new_val));
                    text = new_val;
                    value = parse(text);
                },
                clear: function () {
                    text = '';
                    $input.val('');
                    value = parseFloat();
                },
            };

            return methods;
        }

        var inputs = {};
        inputs.enabled = bool_input('.marker-enabled');
        inputs.freq = numeric_input('.marker-freq', parse_freq);

        $row.find('.marker-del').on('click', function () {
            inputs.enabled.set(false);
            inputs.freq.clear();
        });

        return {
            $: $row,
            inputs: inputs,
            get: function get_marker() {
                var freq_value = inputs.freq.value();
                var snapped = trace_data.amplitude_at_freq(freq_value);

                if (isNaN(snapped.amp)) {
                    $amp.text('');
                }
                else {
                    $amp.text(format_amplitude(snapped.amp, trace_data.units()));
                }

                return {
                    enabled: inputs.enabled.value(),
                    freq: snapped.freq,
                    amp: snapped.amp,
                    number: number,
                };

            },

            get_text: function get_text() {
                return {
                    enabled: inputs.enabled.value(),
                    freq: inputs.freq.text(),
                };
            },

            set: function set_marker(obj) {
                for (var key in inputs) {
                    if (inputs.hasOwnProperty(key) && obj.hasOwnProperty(key)) {
                        inputs[key].set(obj[key]);
                    }
                }
            },
        };
    }

    call(function init_marker_table() {
        var $prototype = $table.find('.marker-table-prototype:first');

        if ($prototype) {
            $table.find('.marker-table-prototype').remove();

            // This gets the outer HTML of the prototype since html()
            // returns the inner HTML only
            $prototype.removeClass('marker-table-prototype');
            var prototype_html = $('<div>').append($prototype.clone()).html();
            $table.data('prototype-html', prototype_html);
        }
        else {
            console.log('No prototype-html for marker_table');
        }

        for (var i = 0; i < num_rows; i++) {
            var elem = $($table.data('prototype-html'));
            elem.addClass('marker-table-row');
            elem.find('.marker-num').text(i+1)
            $table.append(elem)
            rows.push(table_row(elem,i+1));
        }
    });

    function get_markers() {
        return rows.map(function (marker) {
            return marker.get();
        });
    }

    function set_markers(new_markers) {
        for (var i = 0; i < new_markers.length; i++) {
            if (i < rows.length) {
                rows[i].set(new_markers[i]);
            }
        }
    }

    function maybe_place_marker(freq) {
        var $active = $(document.activeElement);

        for (var i = 0; i < rows.length; i++) {
            var row = rows[i];
            if ($.contains(row.$[0], $active[0]) && $active[0].className === 'marker-freq') {
                row.set({
                    freq: freq,
                    enabled: true,
                });
                return row;
            }
        }
    }

    function add_new_marker(freq) {
        // Put this marker in the next empty row in the table
        for (var i = 0; i < rows.length; i++) {
            var row = rows[i];
            if (row.get_text().freq === '') {
                row.set({
                   freq: freq,
                   enabled: true,
                });

                // Make the new marker the active element
                row.$.find('.marker-freq').focus();
                return row;
            }
        }
    }

    return {
        get_markers: get_markers,
        set_markers: set_markers,
        get_marker_text: function () {
            return rows.map(function (marker) {
                return marker.get_text();
            });
        },
        maybe_place_marker: maybe_place_marker,
        add_new_marker: add_new_marker,
    };
}

