
var freq_factors = {
    hz: 1,
    khz: 1e3,
    k: 1e3,
    mhz: 1e6,
    m: 1e6,
    ghz: 1e9,
    g: 1e9,
};

function parse_freq(s) {
    if (typeof s === 'number') return s;
    if (typeof s === 'undefined') return parseFloat();
    s = s.replace(/\s+/g, '');
    var m = s.match(/^([+-]?[0-9]*\.?[0-9]*)\s*([kmg]?(h|hz)?)?$/i);

    if (m) {
        var ret = parseFloat(m[1]);
        if (typeof m[2] === 'string') {
            var factor = freq_factors[m[2].toLowerCase()];
            ret *= factor;
        }
        return ret;
    }

    return parseFloat(); // NaN
}

function parse_amp(s) {
    if (typeof s === 'number') return s;
    if (typeof s === 'undefined') return parseFloat();
    var toks = s.split(/\s+/);
    return parseFloat(toks[0]);
}

function __pure_unit_tester(func, func_name) {
    if (typeof func !== 'function') {
        throw 'Not a function: ' + func;
    }
    func_name = func_name || String(func).substr(0, 60);

    return function pure_unit_tester(input, expected) {
        var actual = func.apply(func, input);

        if (actual !== expected) {
            console.error('Failed unit test for ' + func_name);
            console.error('    Input:    ' + input);
            console.error('    Actual:   ' + actual);
            console.error('    Expected: ' + expected);
            throw 'Failed unit test';
        }
    }
}

var long_format_freq = (function () {
    var freq_units = [
        { factor: 1e9, unit: 'GHz',  digits: 9 },
        { factor: 1e6, unit: 'MHz',  digits: 6 },
        { factor: 1e3, unit: 'kHz',  digits: 3 },
        { factor: 1e0, unit: 'Hz',   digits: 0 },
    ];

    function long_format_freq(hz) {
        hz = Math.round(Number(hz));
        if (isNaN(hz)) return 'NaN';

        var sign = Math.sign(hz);
        hz = sign * hz;

        if (hz < 1e-9) {
        }

        for (var i = 0; i < freq_units.length; i++) {
            var factor = freq_units[i].factor;
            var unit = freq_units[i].unit;
            var digits = freq_units[i].digits;
            if (hz >= factor) {

                var whole = Math.floor(hz / factor);
                var sf = ((hz / factor) - whole).toFixed(digits + 3);
                sf = sf.replace(/^0?\./, '');

                var frac = '';

                for (var k = 0; k < digits; k += 3) {
                    frac += sf.substr(k, 3) + ' ';
                }

                frac = frac.replace(/(000 )+$/g, '').trim();

                var num = whole + '.' + frac;

                num = num.replace(/\.0*$/, '');

                if (num.match(/ [0-9]{1,2}$/)) {
                    throw 'Bad format: number ' + hz + ' -> ' + num;
                }
                return (sign < 0 ? '-' : '') + num + ' ' + unit;
            }
        }

        return '0 Hz';
    }

    var test = __pure_unit_tester(long_format_freq, 'long_format_freq');

    test([0], '0 Hz');
    test([0.1], '0 Hz');
    test([-0.1], '0 Hz');
    test([1], '1 Hz');
    test([-1], '-1 Hz');
    test([999.999999e6], '999.999 999 MHz');
    test([-999.999999e6], '-999.999 999 MHz');

    test([999.999999e6 - 0.100], '999.999 999 MHz');
    test([999.999999e6 - 0.001], '999.999 999 MHz');
    test([999.999999e6 - 0.500], '999.999 999 MHz');
    test([999.999999e6 - 0.600], '999.999 998 MHz');

    test([-999.999999e6 - 0.100], '-999.999 999 MHz');
    test([-999.999999e6 - 0.001], '-999.999 999 MHz');
    test([-999.999999e6 - 0.500], '-999.999 999 MHz');
    test([-999.999999e6 - 0.600], '-1 GHz');

    test([999.999999e6 + 0.100], '999.999 999 MHz');
    test([999.999999e6 + 0.001], '999.999 999 MHz');
    test([999.999999e6 + 0.500], '1 GHz');
    test([999.999999e6 + 0.600], '1 GHz');

    test([-999.999999e6 + 0.100], '-999.999 999 MHz');
    test([-999.999999e6 + 0.001], '-999.999 999 MHz');
    test([-999.999999e6 + 0.500], '-999.999 998 MHz');
    test([-999.999999e6 + 0.600], '-999.999 998 MHz');

    test([1000e6], '1 GHz');
    test([6e9], '6 GHz');

    test([-1000e6], '-1 GHz');
    test([-6e9], '-6 GHz');

    return long_format_freq;
})();

var format_freq = long_format_freq;

(function format_dbm_unit_test() {
    function test(input, expected) {
        var actual = format_amplitude(input, "DBM");
        if (actual !== expected) {
            console.error('Failed format_dbm test!');
            console.error('    Input:    ' + input);
            console.error('    Produced: ' + actual);
            console.error('    Expected: ' + expected);
        }
    }

    test(-1, '-1 dBm');
    test(-0.2, '-0.2 dBm');
    test(-0.03, '-0.03 dBm');
    test(0, '0 dBm');
    test(0.005, '0.01 dBm');
    test(0.06, '0.06 dBm');
    test(0.7, '0.7 dBm');
    test(8, '8 dBm');
    test(-9, '-9 dBm');
})();

function pretty_format_amplitude_unit(desired_units) {
    if (desired_units === 'DBM') {
        return 'dBm';
    }
    else if (desired_units === 'DBUV') {
        return 'dB' + '\u03bc' + 'V';
    }
}

function format_amplitude(amp, desired_units) {
    if (desired_units === 'DBUV') {
        return format_db(amp) + ' ' + pretty_format_amplitude_unit(desired_units);
    }
    else if (desired_units === 'DBM') {
        return format_db(amp) + ' ' + pretty_format_amplitude_unit(desired_units);
    }
    else {
        console.error("Invalid destination units: " + desired_units);
        return '';
    }
}

function format_db(amp) {
    // The instrument has centi-dBm precision
    amp = Number(amp);
    return String(Math.round(amp * 100) / 100);
}

function parse_block_response(response) {
    var value = "";

    if (response && response[0] === '#') {
        var meta_len = +response[1];

        if (meta_len > 0) {
            value = response.slice(2 + meta_len);
        }
    }

    return value;
}

function format_microseconds(value) {
    value = Number(value);
    if (isNaN(value)) {
        return '';
    }
    else if (value < 1) {
        return (value * 1e3) + 'ns'
    }
    else if (value < 1e3) {
        return value + ' us';
    }
    else if (value < 1e6) {
        return (value / 1e3) + ' ms';
    }
    else {
        return (value / 1e6) + ' s';
    }
}

function format_date(d) {
    var months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
    var y = d.getFullYear();
    var m = months[d.getMonth()];
    var d = d.getDate();

    return m + ' ' + d + ', ' + y;
}

function format_time(d) {
    var h = d.getHours();
    var m = String(d.getMinutes());
    if (m.length === 1) m = '0' + m;
    var s = String(d.getSeconds());
    if (s.length === 1) s = '0' + s;
    return h + ':' + m + ':' + s;
}

function format_datetime(d) {
    return format_date(d) + ' ' + format_time(d);
}

function format_bytes(value) {
    value = Number(value);
    if (isNaN(value)) {
        return '';
    }

    if (value < 1024) {
        return value + ' B';
    }

    var units = ['k', 'M', 'G', 'T', 'E'];
    for (var i = 0; (value /= 1024) >= 1024 && i < units.length - 1; ++i)
        ;

    return value.toFixed((value === Math.floor(value) || value >= 100) ?
            0 : ((value >= 10) ? 1 : 2))
            + ' ' + units[i] + 'B';
}
